from fastapi import Request, Response
from starlette.middleware.base import BaseHTTPMiddleware
from starlette.responses import JSONResponse
from starlette.types import ASGIApp

from fastapi_jwks.models.types import JWKSMiddlewareConfig
from fastapi_jwks.validators import JWKSValidator


class JWKSAuthMiddleware(BaseHTTPMiddleware):
    def __init__(
        self,
        app: ASGIApp,
        jwks_validator: JWKSValidator,
        config: JWKSMiddlewareConfig | None = None,
        auth_header: str = "Authorization",
        auth_scheme: str = "Bearer",
        exclude_paths: list[str] | None = None,
    ):
        super().__init__(app)
        self.config = config or JWKSMiddlewareConfig()
        self.jwks_validator = jwks_validator
        self.auth_header = auth_header
        self.auth_scheme = auth_scheme
        self.exclude_paths = [] if exclude_paths is None else exclude_paths

    def unauthorized_response(self) -> JSONResponse:
        return JSONResponse(
            status_code=401,
            content={
                "title": "Unauthorized",
                "detail": "Invalid authorization token",
            },
        )

    async def dispatch(self, request: Request, call_next) -> Response:
        if self.exclude_paths and request.url.path in self.exclude_paths:
            return await call_next(request)

        authorization: str | None = request.headers.get(self.auth_header)
        if not authorization:
            return self.unauthorized_response()

        try:
            scheme, token = authorization.split()
            if scheme.lower() != self.auth_scheme.lower():
                return self.unauthorized_response()
        except ValueError:
            return self.unauthorized_response()

        try:
            payload = self.jwks_validator.validate_token(token)
            setattr(request.state, self.config.payload_field, payload)
            setattr(request.state, self.config.token_field, token)
        except Exception:
            return self.unauthorized_response()

        return await call_next(request)



from datetime import timedelta
from typing import Any

from pydantic import BaseModel


class JWTDecodeConfig(BaseModel):
    audience: list[str] | None = None
    issuer: str | None = None
    leeway: float | timedelta | None = None
    options: dict[str, Any] | None = None
    verify: bool | None = None


class JWKSMiddlewareConfig(BaseModel):
    payload_field: str = "payload"
    token_field: str = "raw_token"


class JWKSConfig(BaseModel):
    url: str
    ca_cert_path: str | None = None


class JWTTokenInjectorConfig(BaseModel):
    payload_field: str = "payload"
    token_field: str = "raw_token"



import logging
from functools import cached_property
from typing import Any, Generic, TypeVar

import httpx
import jwt
from cachetools import TTLCache, cached
from fastapi import HTTPException
from jwt import algorithms
from pydantic import BaseModel

from fastapi_jwks.models.types import JWKSConfig, JWTDecodeConfig

logger = logging.getLogger("fastapi-jwks")

DataT = TypeVar("DataT", bound=BaseModel)


class JWKSValidator(Generic[DataT]):
    def __init__(self, decode_config: JWTDecodeConfig, jwks_config: JWKSConfig):
        self.decode_config = decode_config
        self.jwks_config = jwks_config
        self.client = self._create_client()

    def _create_client(self) -> httpx.Client:
        client_kwargs = {}
        if self.jwks_config.ca_cert_path:
            client_kwargs["verify"] = self.jwks_config.ca_cert_path
        return httpx.Client(**client_kwargs)

    @cached(cache=TTLCache(ttl=600, maxsize=1))
    def jwks_data(self) -> dict[str, Any]:
        try:
            logger.debug("Fetching JWKS from %s", self.jwks_config.url)
            jwks_response = self.client.get(self.jwks_config.url)
            jwks_response.raise_for_status()
        except httpx.RequestError as e:
            raise HTTPException(status_code=503, detail="Invalid JWKS URI") from e
        jwks: dict[str, Any] = jwks_response.json()
        if "keys" not in jwks:
            raise HTTPException(status_code=503, detail="Invalid JWKS")
        return jwks

    @staticmethod
    def __extract_algorithms(jwks_response: dict[str, Any]) -> list[str]:
        return [key["alg"] for key in jwks_response["keys"]]

    @cached_property
    def __is_generic_passed(self):
        if getattr(self, "__orig_class__", None) is None:  # type: ignore
            return False
        return True

    def validate_token(self, token: str) -> DataT:
        if not self.__is_generic_passed:
            raise ValueError(
                "Validator needs a model as generic value to decode payload"
            )

        public_key = None
        try:
            header = jwt.get_unverified_header(token)
            kid = header["kid"]
            jwks_data = self.jwks_data()
            provided_algorithms = self.__extract_algorithms(jwks_data)
            if header["alg"] not in provided_algorithms:
                logger.debug(
                    f"Could not find '{header['alg']}' in provided algorithms: {provided_algorithms}"
                )
                raise HTTPException(status_code=401, detail="Invalid token")
            for key in jwks_data["keys"]:
                if key["kid"] == kid:
                    public_key = algorithms.get_default_algorithms()[
                        header["alg"]
                    ].from_jwk(key)
                    break
            if public_key is None:
                logger.debug(
                    f"No public key for provided algorithm '{header['alg']}' found in JWKS data"
                )
                raise HTTPException(status_code=401, detail="Invalid token")
            return self.__orig_class__.__args__[0].model_validate(  # type: ignore
                # This line gets the generic value in runtime to transform it to the correct pydantic model
                jwt.decode(
                    token,
                    key=public_key,
                    **self.decode_config.model_dump(),
                    algorithms=[header["alg"]],
                )
            )
        except jwt.ExpiredSignatureError:
            logger.debug("Expired token", exc_info=True)
            raise HTTPException(status_code=401, detail="Token has expired") from None
        except jwt.InvalidTokenError:
            logger.debug("Invalid token", exc_info=True)
            raise HTTPException(status_code=401, detail="Invalid token") from None



from typing import Generic, TypeVar

from pydantic import BaseModel
from starlette.requests import Request

from fastapi_jwks.models.types import JWTTokenInjectorConfig

TypeT = TypeVar("TypeT", bound=BaseModel)


class JWTTokenInjector(Generic[TypeT]):
    def __init__(self, config: JWTTokenInjectorConfig | None = None):
        self.config = config or JWTTokenInjectorConfig()

    async def __call__(self, request: Request) -> TypeT:
        return getattr(request.state, self.config.payload_field)


class JWTRawTokenInjector:
    def __init__(self, config: JWTTokenInjectorConfig | None = None):
        self.config = config or JWTTokenInjectorConfig()

    async def __call__(self, request: Request) -> str:
        return getattr(request.state, self.config.token_field)
